@@ -12,7 +12,6 @@
 # License for the specific language.
 """Module for InfidelityNode."""
 from typing import (
-    Dict,
     List,
     Optional,
 )
@@ -20,52 +19,22 @@ from typing import (
 import forge
 import numpy as np
 
-from qctrlcommons.node.module import ModuleNode
-from qctrlcommons.node.utils import validate_values_and_batch_shape
+from qctrlcommons.node import types
+from qctrlcommons.node.base import Node
+from qctrlcommons.node.node_data import (
+    StfNodeData,
+    TargetNodeData,
+    TensorNodeData,
+    TensorPwcNodeData,
+)
+from qctrlcommons.node.utils import validate_batch_and_values_shapes
 from qctrlcommons.preconditions import (
     check_argument,
     check_sample_times,
 )
 
-from . import types
-from .node_data import (
-    StfNodeData,
-    TargetNodeData,
-    TensorPwcNodeData,
-)
-from .tensorflow import TensorNodeData
 
-
-class InfidelityNode(ModuleNode):
-    """Represent InfidelityNode."""
-
-    _module_name = "qctrlcore"
-
-    def _evaluate_node(
-        self, execution_context: "ExecutionContext", args: List, kwargs: Dict
-    ):
-        """Handle the condition that cannot use target as kwargs in pythonflow.
-
-        Parameters
-        ----------
-        execution_context : ExecutionContext
-            helper class for evaluating the value of the node.
-        args : List
-            argument list for the node.
-        kwargs : Dict
-            keyword arguments for the node.
-
-        Returns
-        -------
-        Any
-            the value of the node.
-        """
-        if kwargs.get("target_operator", None):
-            kwargs["target"] = kwargs.pop("target_operator")
-        return super()._evaluate_node(execution_context, args, kwargs)
-
-
-class InfidelityPwc(InfidelityNode):
+class InfidelityPwc(Node):
     r"""
     Creates the total infidelity of the given piecewise-constant system.
 
@@ -117,7 +86,10 @@ class InfidelityPwc(InfidelityNode):
 
     See Also
     --------
-    infidelity_stf : Corresponding function for sampleable controls.
+    infidelity_stf : Corresponding operation for `Stf` Hamiltonians.
+    time_evolution_operators_pwc : Unitary time evolution operators for quantum systems with
+        `TensorPwc` Hamiltonians.
+    target : Define the target operation of the time evolution.
 
     Notes
     -----
@@ -164,7 +136,6 @@ class InfidelityPwc(InfidelityNode):
     """
 
     name = "infidelity_pwc"
-    _module_attr = "create_infidelity"
     args = [
         forge.arg("hamiltonian", type=types.TensorPwc),
         forge.arg("target_operator", type=types.Target),
@@ -189,7 +160,7 @@ class InfidelityPwc(InfidelityNode):
             "The target_operator must be a Target.",
             {"target_operator": target_operator},
         )
-        _, batch_shape = validate_values_and_batch_shape(hamiltonian, "hamiltonian")
+        batch_shape, _ = validate_batch_and_values_shapes(hamiltonian, "hamiltonian")
         if noise_operators is not None:
             check_argument(
                 len(noise_operators) == 0 or batch_shape == (),
@@ -197,18 +168,16 @@ class InfidelityPwc(InfidelityNode):
                 {"hamiltonian": hamiltonian, "noise_operators": noise_operators},
             )
             check_argument(
-                all(
-                    [isinstance(noise, TensorPwcNodeData) for noise in noise_operators]
-                ),
+                all(isinstance(noise, TensorPwcNodeData) for noise in noise_operators),
                 "Each of the noise_operators must be a TensorPwc.",
                 {"noise_operators": noise_operators},
             )
         return TensorNodeData(_operation, shape=batch_shape)
 
 
-class InfidelityStf(InfidelityNode):
+class InfidelityStf(Node):
     r"""
-    Creates the total infidelity of the given system.
+    Creates the total infidelity of a given system with a sampleable Hamiltonian.
 
     See :obj:`infidelity_pwc` for information about the total infidelity
     created by this function.
@@ -242,11 +211,13 @@ class InfidelityStf(InfidelityNode):
 
     See Also
     --------
-    infidelity_pwc : Corresponding function for piecewise-constant controls.
+    infidelity_pwc : Corresponding operation for `TensorPwc` Hamiltonians.
+    target : Define the target operation of the time evolution.
+    time_evolution_operators_stf : Unitary time evolution operators for quantum systems with
+        `Stf` Hamiltonians.
     """
 
     name = "infidelity_stf"
-    _module_attr = "create_infidelity_stf"
     args = [
         forge.arg("sample_times", type=np.ndarray),
         forge.arg("hamiltonian", type=types.Stf),
@@ -279,9 +250,9 @@ class InfidelityStf(InfidelityNode):
         )
         if noise_operators is not None:
             check_argument(
-                all([isinstance(noise, StfNodeData) for noise in noise_operators]),
+                all(isinstance(noise, StfNodeData) for noise in noise_operators),
                 "Each of the noise_operators must be an Stf.",
                 {"noise_operators": noise_operators},
             )
-        _, batch_shape = validate_values_and_batch_shape(hamiltonian, "hamiltonian")
+        batch_shape, _ = validate_batch_and_values_shapes(hamiltonian, "hamiltonian")
         return TensorNodeData(_operation, shape=batch_shape)
