@@ -20,9 +20,11 @@ import forge
 import numpy as np
 
 from qctrlcommons.node import types
-from qctrlcommons.node.module import CoreNode
-from qctrlcommons.node.node_data import TensorPwcNodeData
-from qctrlcommons.node.tensorflow import TensorNodeData
+from qctrlcommons.node.base import Node
+from qctrlcommons.node.node_data import (
+    TensorNodeData,
+    TensorPwcNodeData,
+)
 from qctrlcommons.node.utils import (
     validate_ms_shapes,
     validate_shape,
@@ -35,7 +37,7 @@ from qctrlcommons.preconditions import (
 )
 
 
-class MsPhases(CoreNode):
+class MsPhases(Node):
     r"""
     Calculates the relative phases for all pairs of ions described by a Mølmer–Sørensen-type
     interaction.
@@ -81,7 +83,9 @@ class MsPhases(CoreNode):
 
     See Also
     --------
-    :func:`~qctrl.dynamic.namespaces.FunctionNamespace.calculate_ion_chain_properties`
+    :func:`~qctrl.dynamic.namespaces.FunctionNamespace.calculate_ion_chain_properties` : Calculate
+        the properties of an ion chain.
+    ms_infidelity : Final operational infidelity of a Mølmer–Sørensen gate.
 
     Notes
     -----
@@ -131,12 +135,11 @@ class MsPhases(CoreNode):
     References
     ----------
     .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
-            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies, 202000044 (2020).
+            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies 3, 2000044 (2020).
             <https://doi.org/10.1002/qute.202000044>`_
     """
 
     name = "ms_phases"
-    _module_attr = "ions.ms_phases"
     args = [
         forge.arg("drives", type=List[types.TensorPwc]),
         forge.arg("lamb_dicke_parameters", type=np.ndarray),
@@ -153,7 +156,7 @@ class MsPhases(CoreNode):
         relative_detunings = kwargs.get("relative_detunings")
         check_argument_iteratable(drives, "drives")
         check_argument(
-            all([isinstance(drive, TensorPwcNodeData) for drive in drives]),
+            all(isinstance(drive, TensorPwcNodeData) for drive in drives),
             "Each of the drives must be a TensorPwc.",
             {"drives": drives},
         )
@@ -173,7 +176,7 @@ class MsPhases(CoreNode):
         return TensorNodeData(_operation, shape=shape)
 
 
-class MsDisplacements(CoreNode):
+class MsDisplacements(Node):
     r"""
     Calculates the displacements for each mode and ion combination where ions are described by
     a Mølmer–Sørensen-type interaction.
@@ -218,7 +221,9 @@ class MsDisplacements(CoreNode):
 
     See Also
     --------
-    :func:`~qctrl.dynamic.namespaces.FunctionNamespace.calculate_ion_chain_properties`
+    :func:`~qctrl.dynamic.namespaces.FunctionNamespace.calculate_ion_chain_properties` : Calculate
+        the properties of an ion chain.
+    ms_infidelity : Final operational infidelity of a Mølmer–Sørensen gate.
 
     Notes
     -----
@@ -275,12 +280,11 @@ class MsDisplacements(CoreNode):
     References
     ----------
     .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
-            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies, 202000044 (2020).
+            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies 3, 2000044 (2020).
             <https://doi.org/10.1002/qute.202000044>`_
     """
 
     name = "ms_displacements"
-    _module_attr = "ions.ms_displacements"
     args = [
         forge.arg("drives", type=List[types.TensorPwc]),
         forge.arg("lamb_dicke_parameters", type=np.ndarray),
@@ -297,7 +301,7 @@ class MsDisplacements(CoreNode):
         relative_detunings = kwargs.get("relative_detunings")
         check_argument_iteratable(drives, "drives")
         check_argument(
-            all([isinstance(drive, TensorPwcNodeData) for drive in drives]),
+            all(isinstance(drive, TensorPwcNodeData) for drive in drives),
             "Each of the drives must be a TensorPwc.",
             {"drives": drives},
         )
@@ -317,7 +321,7 @@ class MsDisplacements(CoreNode):
         return TensorNodeData(_operation, shape=shape)
 
 
-class MsInfidelity(CoreNode):
+class MsInfidelity(Node):
     r"""
     Calculates the final operational infidelity of the Mølmer–Sørensen gate.
 
@@ -360,7 +364,9 @@ class MsInfidelity(CoreNode):
 
     See Also
     --------
-    ms_displacements, ms_phases
+    ms_dephasing_robust_cost : Cost for robust optimization of a Mølmer–Sørensen gate.
+    ms_displacements : Displacements for each mode/ion combination.
+    ms_phases : Relative phases for all pairs of ions.
 
     Notes
     -----
@@ -376,12 +382,11 @@ class MsInfidelity(CoreNode):
     References
     ----------
     .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
-            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies, 202000044 (2020).
+            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies 3, 2000044 (2020).
             <https://doi.org/10.1002/qute.202000044>`_
     """
 
     name = "ms_infidelity"
-    _module_attr = "ions.ms_infidelity"
     args = [
         forge.arg("phases", type=types.Tensor),
         forge.arg("displacements", type=types.Tensor),
@@ -460,7 +465,7 @@ class MsInfidelity(CoreNode):
         return TensorNodeData(_operation, shape=shape)
 
 
-class MsDephasingRobustCost(CoreNode):
+class MsDephasingRobustCost(Node):
     r"""
     Calculates the cost for robust optimization of a Mølmer–Sørensen gate.
 
@@ -505,8 +510,9 @@ class MsDephasingRobustCost(CoreNode):
 
     See Also
     --------
-    :func:`~qctrl.dynamic.namespaces.FunctionNamespace.calculate_ion_chain_properties`
-    ms_infidelity
+    :func:`~qctrl.dynamic.namespaces.FunctionNamespace.calculate_ion_chain_properties` : Calculate
+        the properties of an ion chain.
+    ms_infidelity : Final operational infidelity of a Mølmer–Sørensen gate.
 
     Notes
     -----
@@ -553,11 +559,10 @@ class MsDephasingRobustCost(CoreNode):
             M. J. Biercuk, Physical Review Applied 13, 024022 (2020).
             <https://doi.org/10.1103/PhysRevApplied.13.024022>`_
     .. [2] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
-            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies, 2000044 (2020).
+            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies 3, 2000044 (2020).
             <https://doi.org/10.1002/qute.202000044>`_
     """
     name = "ms_dephasing_robust_cost"
-    _module_attr = "ions.ms_dephasing_robust_cost"
     args = [
         forge.arg("drives", type=List[types.TensorPwc]),
         forge.arg("lamb_dicke_parameters", type=np.ndarray),
@@ -572,7 +577,7 @@ class MsDephasingRobustCost(CoreNode):
         relative_detunings = kwargs.get("relative_detunings")
         check_argument_iteratable(drives, "drives")
         check_argument(
-            all([isinstance(drive, TensorPwcNodeData) for drive in drives]),
+            all(isinstance(drive, TensorPwcNodeData) for drive in drives),
             "Each of the drives must be a TensorPwc.",
             {"drives": drives},
         )
