@@ -12,8 +12,6 @@
 # License for the specific language.
 """Module for all the node related to PWC."""
 from typing import (
-    Any,
-    Dict,
     List,
     Optional,
     Union,
@@ -23,13 +21,15 @@ import forge
 import numpy as np
 
 from qctrlcommons.node import types
-from qctrlcommons.node.module import CoreNode
-from qctrlcommons.node.node_data import TensorPwcNodeData
-from qctrlcommons.node.tensorflow import TensorNodeData
+from qctrlcommons.node.base import Node
+from qctrlcommons.node.node_data import (
+    TensorNodeData,
+    TensorPwcNodeData,
+)
 from qctrlcommons.node.utils import (
+    validate_batch_and_values_shapes,
     validate_hamiltonian,
     validate_shape,
-    validate_values_and_batch_shape,
 )
 from qctrlcommons.preconditions import (
     check_argument,
@@ -37,11 +37,12 @@ from qctrlcommons.preconditions import (
     check_argument_iteratable,
     check_argument_numeric,
     check_duration,
+    check_operator,
     check_sample_times,
 )
 
 
-class TensorPwc(CoreNode):
+class TensorPwc(Node):
     r"""
     Creates a piecewise-constant function of time.
 
@@ -78,10 +79,14 @@ class TensorPwc(CoreNode):
         functions :math:`v(t)`, each of shape
         :math:`D_1 \times \ldots \times D_m`.
 
+    See Also
+    --------
+    pwc_operator : Create `TensorPwc` operators.
+    pwc_signal : Create `TensorPwc` signals from (possibly complex) values.
+    pwc_sum : Sum multiple `TensorPwc`s.
     """
 
     name = "tensor_pwc"
-    _module_attr = "TensorPwc"
     args = [
         forge.arg("durations", type=np.ndarray),
         forge.arg("values", type=Union[np.ndarray, types.Tensor]),
@@ -113,6 +118,8 @@ class TensorPwc(CoreNode):
         )
         for index, duration in enumerate(durations):
             check_duration(duration, f"durations[{index}]")
+        if time_dimension < 0:
+            time_dimension = len(values_shape) + time_dimension
         return TensorPwcNodeData(
             _operation,
             values_shape=values_shape[time_dimension + 1 :],
@@ -121,7 +128,7 @@ class TensorPwc(CoreNode):
         )
 
 
-class PwcSignal(CoreNode):
+class PwcSignal(Node):
     r"""
     Creates a piecewise-constant signal (scalar-valued function of time).
 
@@ -154,12 +161,14 @@ class PwcSignal(CoreNode):
 
     See Also
     --------
-    tensor_pwc : Corresponding function with support for segments of different
-        durations.
+    complex_pwc_signal : Create complex `TensorPwc` signals from their moduli and phases.
+    pwc_operator : Create `TensorPwc` operators.
+    pwc_sum : Sum multiple `TensorPwc`s.
+    symmetrize_pwc : Symmetrize `TensorPwc`s.
+    tensor_pwc : Corresponding operation with support for segments of different durations.
     """
 
     name = "pwc_signal"
-    _module_attr = "create_pwc_signal"
     args = [
         forge.arg("values", type=Union[np.ndarray, types.Tensor]),
         forge.arg("duration", type=float),
@@ -187,7 +196,7 @@ class PwcSignal(CoreNode):
         )
 
 
-class ComplexPwcSignal(CoreNode):
+class ComplexPwcSignal(Node):
     r"""
     Creates a complex piecewise-constant signal from moduli and phases.
 
@@ -226,12 +235,10 @@ class ComplexPwcSignal(CoreNode):
 
     See Also
     --------
-    pwc_signal : Corresponding function for creating signals from
-        (possibly-complex) values.
+    pwc_signal : Create `TensorPwc` signals from (possibly complex) values.
     """
 
     name = "complex_pwc_signal"
-    _module_attr = "create_complex_pwc_signal"
     args = [
         forge.arg("moduli", type=Union[np.ndarray, types.Tensor]),
         forge.arg("phases", type=Union[np.ndarray, types.Tensor]),
@@ -268,7 +275,196 @@ class ComplexPwcSignal(CoreNode):
         )
 
 
-class PwcOperatorHermitianPart(CoreNode):
+class PwcOperator(Node):
+    """
+    Creates a constant operator multiplied by a piecewise-constant signal.
+
+    Parameters
+    ----------
+    signal : TensorPwc
+        The piecewise-constant signal :math:`a(t)`, or a batch of
+        piecewise-constant signals.
+    operator : np.ndarray or Tensor
+        The operator :math:`A`. It must have two equal dimensions.
+    name : str, optional
+        The name of the node.
+
+    Returns
+    -------
+    TensorPwc
+        The piecewise-constant operator :math:`a(t)A` (or a batch of
+        piecewise-constant operators, if you provide a batch of
+        piecewise-constant signals).
+
+    See Also
+    --------
+    complex_pwc_signal : Create complex `TensorPwc` signals from their moduli and phases.
+    constant_pwc_operator : Create constant `TensorPwc`s.
+    pwc_operator_hermitian_part : Hermitian part of a piecewise-constant operator.
+    pwc_signal : Create `TensorPwc` signals from (possibly complex) values.
+    pwc_sum : Sum multiple `TensorPwc`s.
+    stf_operator : Corresponding operation for `Stf`s.
+    tensor_pwc : Create piecewise-constant functions.
+    """
+
+    name = "pwc_operator"
+    args = [
+        forge.arg("signal", type=types.TensorPwc),
+        forge.arg("operator", type=Union[np.ndarray, types.Tensor]),
+    ]
+    rtype = types.TensorPwc
+
+    @classmethod
+    def create_node_data(cls, _operation, **kwargs):
+        signal = kwargs.get("signal")
+        operator = kwargs.get("operator")
+        check_argument(
+            isinstance(signal, TensorPwcNodeData),
+            "The signal must be a TensorPwc.",
+            {"signal": signal},
+        )
+        batch_shape, _ = validate_batch_and_values_shapes(signal, "signal")
+        values_shape = validate_shape(operator, "operator")
+        check_operator(operator, "operator")
+        check_argument(
+            len(values_shape) == 2,
+            "The operator must be a matrix, not a batch.",
+            {"operator": operator},
+            extras={"operator.shape": values_shape},
+        )
+        return TensorPwcNodeData(
+            _operation,
+            values_shape=values_shape,
+            durations=signal.durations,
+            batch_shape=batch_shape,
+        )
+
+
+class ConstantPwcOperator(Node):
+    r"""
+    Creates a constant piecewise-constant operator over a specified duration.
+
+    Parameters
+    ----------
+    duration : float
+        The duration :math:`\tau` for the resulting piecewise-constant
+        operator.
+    operator : np.ndarray or Tensor
+        The operator :math:`A`, or a batch of operators. It must have at
+        least two dimensions, and its last two dimensions must be equal.
+    name : str, optional
+        The name of the node.
+
+    Returns
+    -------
+    TensorPwc
+        The constant operator :math:`t\mapsto A` (for :math:`0\leq t\leq\tau`)
+        (or a batch of constant operators, if you provide a batch of operators).
+
+    See Also
+    --------
+    constant_stf_operator : Corresponding operation for `Stf`s.
+    pwc_operator : Create `TensorPwc` operators.
+    pwc_operator_hermitian_part : Hermitian part of a piecewise-constant operator.
+    """
+
+    name = "constant_pwc_operator"
+    args = [
+        forge.arg("duration", type=float),
+        forge.arg("operator", type=Union[np.ndarray, types.Tensor]),
+    ]
+    rtype = types.TensorPwc
+
+    @classmethod
+    def create_node_data(cls, _operation, **kwargs):
+        duration = kwargs.get("duration")
+        operator = kwargs.get("operator")
+        check_duration(duration, "duration")
+        shape = validate_shape(operator, "operator")
+        check_operator(operator, "operator")
+        return TensorPwcNodeData(
+            _operation,
+            values_shape=shape[-2:],
+            durations=np.array([duration]),
+            batch_shape=shape[:-2],
+        )
+
+
+class ConstantPwc(Node):
+    r"""
+    Creates a piecewise-constant (PWC) function of time that is constant over a specified duration.
+
+    Parameters
+    ----------
+    constant : np.ndarray or Tensor
+        The value :math:`c` of the function on the constant segment.
+        To create a batch of :math:`B_1 \times \ldots \times B_n` piecewise-constant
+        functions of shape :math:`D_1 \times \ldots \times D_m`, provide this `constant`
+        parameter as an object of shape
+        :math:`B_1\times\ldots\times B_n\times D_1\times\ldots\times D_m`.
+    duration : float
+        The duration :math:`\tau` for the resulting piecewise-constant function.
+    batch_dimension_count : int, optional
+        The number of batch dimensions, :math:`n` in `constant`.
+        If provided, the first :math:`n` dimensions of `constant` are considered batch dimensions.
+        Defaults to 0, which corresponds to no batch.
+    name : str, optional
+        The name of the node.
+
+    Returns
+    -------
+    TensorPwc
+        The constant function :math:`f(t) = c` (for :math:`0\leq t\leq\tau`)
+        (or a batch of constant functions, if you provide `batch_dimension_count`).
+
+    See Also
+    --------
+    constant_pwc_operator : Create constant `TensorPwc` operators.
+    constant_stf : Corresponding operation for `Stf`s.
+    tensor_pwc : Create piecewise-constant functions.
+    """
+
+    name = "constant_pwc"
+    args = [
+        forge.arg("constant", type=Union[np.ndarray, types.Tensor]),
+        forge.arg("duration", type=float),
+        forge.arg("batch_dimension_count", type=int, default=0),
+    ]
+    rtype = types.TensorPwc
+
+    @classmethod
+    def create_node_data(cls, _operation, **kwargs):
+        constant = kwargs.get("constant")
+        duration = kwargs.get("duration")
+        batch_dimension_count = kwargs.get("batch_dimension_count")
+
+        check_argument_numeric(constant, "constant")
+        check_duration(duration, "duration")
+        check_argument_integer(batch_dimension_count, "batch_dimension_count")
+
+        shape = validate_shape(constant, "constant")
+        check_argument(
+            len(shape) >= batch_dimension_count,
+            "Number of batch dimensions must not be larger than number of dimensions "
+            "of the input constant.",
+            {"constant": constant, "batch_dimension_count": batch_dimension_count},
+            {"Number of value dimensions": len(constant.shape)},
+        )
+        check_argument(
+            batch_dimension_count >= 0,
+            "Number of batch dimensions must not be negative.",
+            {"batch_dimension_count": batch_dimension_count},
+        )
+
+        return TensorPwcNodeData(
+            _operation,
+            values_shape=shape[batch_dimension_count:],
+            durations=np.array([duration]),
+            batch_shape=shape[:batch_dimension_count],
+        )
+
+
+class PwcOperatorHermitianPart(Node):
     r"""
     Creates the Hermitian part of a piecewise-constant operator.
 
@@ -283,10 +479,14 @@ class PwcOperatorHermitianPart(CoreNode):
     -------
     TensorPwc(3D)
         The Hermitian part :math:`\frac{1}{2}(A(t)+A^\dagger(t))`.
+
+    See Also
+    --------
+    constant_pwc_operator : Create constant `TensorPwc`s.
+    pwc_operator : Create `TensorPwc` operators.
     """
 
     name = "pwc_operator_hermitian_part"
-    _module_attr = "get_pwc_operator_hermitian_part"
     args = [
         forge.arg("operator", type=types.TensorPwc),
     ]
@@ -300,7 +500,7 @@ class PwcOperatorHermitianPart(CoreNode):
             "The operator must be a TensorPwc.",
             {"operator": operator},
         )
-        values_shape, batch_shape = validate_values_and_batch_shape(
+        batch_shape, values_shape = validate_batch_and_values_shapes(
             operator,
             "operator",
         )
@@ -312,7 +512,7 @@ class PwcOperatorHermitianPart(CoreNode):
         )
 
 
-class PwcSum(CoreNode):
+class PwcSum(Node):
     r"""
     Creates the sum of multiple piecewise-constant terms.
 
@@ -322,7 +522,9 @@ class PwcSum(CoreNode):
         The individual piecewise-constant terms :math:`\{v_j(t)\}` to sum. All
         terms must have the same duration, values of the same shape, and the
         same batch shape, but may have different segmentations (different
-        numbers of segments of different durations).
+        numbers of segments of different durations) and different dtypes of
+        segment values.
+
     name : str, optional
         The name of the node.
 
@@ -333,11 +535,19 @@ class PwcSum(CoreNode):
         :math:`\sum_j v_j(t)`. Its values have the same shape as the values of
         each of the `terms` that you provided. If each of the `terms` represents
         a batch of functions, this result represents a batch of functions with
-        the same batch shape.
+        the same batch shape. If any `term` has complex-valued segments, the
+        value of the returned TensorPwc is complex, otherwise is float.
+
+    See Also
+    --------
+    discretize_stf : Discretize an `Stf` into a `TensorPwc`.
+    pwc_operator : Create `TensorPwc` operators.
+    pwc_signal : Create `TensorPwc` signals from (possibly complex) values.
+    stf_sum : Corresponding operation for `Stf`s.
+    tensor_pwc : Create piecewise-constant functions.
     """
 
     name = "pwc_sum"
-    _module_attr = "get_pwc_sum"
     args = [forge.arg("terms", type=List[types.TensorPwc])]
     rtype = types.TensorPwc
 
@@ -346,33 +556,29 @@ class PwcSum(CoreNode):
         terms = kwargs.get("terms")
         check_argument_iteratable(terms, "terms")
         check_argument(
-            all([isinstance(term, TensorPwcNodeData) for term in terms]),
+            all(isinstance(term, TensorPwcNodeData) for term in terms),
             "Each of the terms must be a TensorPwc.",
             {"terms": terms},
         )
-        values_shape, batch_shape = validate_values_and_batch_shape(
+        batch_shape, values_shape = validate_batch_and_values_shapes(
             terms[0],
             "terms[0]",
         )
         check_argument(
             all(
-                [
-                    (
-                        (term.values_shape == values_shape)
-                        and (term.batch_shape == batch_shape)
-                    )
-                    for term in terms[1:]
-                ]
+                (
+                    (term.values_shape == values_shape)
+                    and (term.batch_shape == batch_shape)
+                )
+                for term in terms[1:]
             ),
             "All the terms must have the same shape.",
             {"terms": terms},
         )
         check_argument(
             all(
-                [
-                    np.allclose(np.sum(terms[0].durations), np.sum(term.durations))
-                    for term in terms
-                ]
+                np.allclose(np.sum(terms[0].durations), np.sum(term.durations))
+                for term in terms
             ),
             "All the terms must have the same total duration.",
             {"terms": terms},
@@ -399,7 +605,7 @@ class PwcSum(CoreNode):
         )
 
 
-class RealFourierPwcSignal(CoreNode):
+class RealFourierPwcSignal(Node):
     r"""
     Creates a piecewise-constant signal constructed from Fourier components.
 
@@ -488,7 +694,6 @@ class RealFourierPwcSignal(CoreNode):
 
     name = "real_fourier_pwc_signal"
     optimizable_variable = True
-    _module_attr = "create_real_fourier_pwc_signal"
     args = [
         forge.arg("duration", type=float),
         forge.arg("segments_count", type=int),
@@ -500,29 +705,6 @@ class RealFourierPwcSignal(CoreNode):
     ]
     rtype = types.TensorPwc
 
-    def _evaluate_node(
-        self, execution_context: "ExecutionContext", args: List, kwargs: Dict
-    ) -> Any:
-        """Add variable factory to args.
-
-        Parameters
-        ----------
-        execution_context : ExecutionContext
-            helper class for evaluating the value of the node.
-        args: List
-            argument list for the node.
-        kwargs: Dict
-            keyword arguments for the node.
-
-        Returns
-        -------
-        Any
-            the value of the node.
-        """
-        variable_factory = execution_context.variable_factory
-        args.insert(0, variable_factory)
-        return super()._evaluate_node(execution_context, args, kwargs)
-
     @classmethod
     def create_node_data(cls, _operation, **kwargs):
         duration = kwargs.get("duration")
@@ -531,10 +713,30 @@ class RealFourierPwcSignal(CoreNode):
         check_argument_integer(segments_count, "segments_count")
         check_argument(
             segments_count > 0,
-            "The segments count must be greater than zero.",
+            "The number of segments must be greater than zero.",
             {"segments_count": segments_count},
         )
         durations = duration / segments_count * np.ones(segments_count)
+        optimizable_frequencies_count = kwargs.get("optimizable_frequencies_count")
+        if optimizable_frequencies_count is not None:
+            check_argument_integer(
+                optimizable_frequencies_count, "optimizable_frequencies_count"
+            )
+            check_argument(
+                optimizable_frequencies_count > 0,
+                "The number of optimizable frequencies (if provided) must be greater than zero.",
+                {"optimizable_frequencies_count": optimizable_frequencies_count},
+            )
+        randomized_frequencies_count = kwargs.get("randomized_frequencies_count")
+        if randomized_frequencies_count is not None:
+            check_argument_integer(
+                randomized_frequencies_count, "randomized_frequencies_count"
+            )
+            check_argument(
+                randomized_frequencies_count > 0,
+                "The number of randomized frequencies (if provided) must be greater than zero.",
+                {"randomized_frequencies_count": randomized_frequencies_count},
+            )
         return TensorPwcNodeData(
             _operation,
             values_shape=(),
@@ -543,7 +745,7 @@ class RealFourierPwcSignal(CoreNode):
         )
 
 
-class SymmetrizePwc(CoreNode):
+class SymmetrizePwc(Node):
     r"""
     Creates the symmetrization of a piecewise-constant function.
 
@@ -561,10 +763,13 @@ class SymmetrizePwc(CoreNode):
         :math:`w(t)=v(t)` for :math:`0\leq t\leq \tau` and
         :math:`w(t)=v(2\tau-t)` for :math:`\tau\leq t\leq 2\tau`, where
         :math:`\tau` is the duration of :math:`v(t)`.
+
+    See Also
+    --------
+    pwc_signal : Create `TensorPwc` signals from (possibly complex) values.
     """
 
     name = "symmetrize_pwc"
-    _module_attr = "symmetrize_pwc"
     args = [forge.arg("pwc", type=types.TensorPwc)]
     rtype = types.TensorPwc
 
@@ -576,7 +781,7 @@ class SymmetrizePwc(CoreNode):
             "The pwc parameter must be a TensorPwc.",
             {"pwc": pwc},
         )
-        values_shape, batch_shape = validate_values_and_batch_shape(
+        batch_shape, values_shape = validate_batch_and_values_shapes(
             pwc,
             "pwc",
         )
@@ -589,7 +794,7 @@ class SymmetrizePwc(CoreNode):
         )
 
 
-class TimeEvolutionOperatorsPwc(CoreNode):
+class TimeEvolutionOperatorsPwc(Node):
     """
     Calculates the unitary time-evolution operators for a system defined by a piecewise-constant
     Hamiltonian.
@@ -610,10 +815,14 @@ class TimeEvolutionOperatorsPwc(CoreNode):
         Tensor of shape [..., N, D, D], representing the unitary time evolution.
         The n-th element (along the -3 dimension) represents the unitary (or batch of unitaries)
         from t = 0 to ``sample_times[n]``.
+
+    See Also
+    --------
+    state_evolution_pwc : Evolve a quantum state.
+    time_evolution_operators_stf : Corresponding operation for `Stf` Hamiltonians.
     """
 
     name = "time_evolution_operators_pwc"
-    _module_attr = "create_unitaries_from_pwc_hamiltonian"
     args = [
         forge.arg("hamiltonian", type=types.TensorPwc),
         forge.arg("sample_times", type=np.ndarray),
@@ -629,7 +838,7 @@ class TimeEvolutionOperatorsPwc(CoreNode):
             "The Hamiltonian must be a TensorPwc.",
             {"hamiltonian": hamiltonian},
         )
-        values_shape, batch_shape = validate_values_and_batch_shape(
+        batch_shape, values_shape = validate_batch_and_values_shapes(
             hamiltonian,
             "hamiltonian",
         )
