@@ -11,6 +11,7 @@
 # distributed under the License is distributed on an "AS IS" BASIS. See the
 # License for the specific language.
 """Module for sparse."""
+import warnings
 from typing import (
     List,
     Optional,
@@ -19,18 +20,15 @@ from typing import (
 
 import forge
 import numpy as np
-from scipy.sparse import (
-    coo_matrix,
-    issparse,
-)
+from scipy.sparse import coo_matrix
 
 from qctrlcommons.node import types
-from qctrlcommons.node.module import CoreNode
+from qctrlcommons.node.base import Node
 from qctrlcommons.node.node_data import (
     SparsePwcNodeData,
+    TensorNodeData,
     TensorPwcNodeData,
 )
-from qctrlcommons.node.tensorflow import TensorNodeData
 from qctrlcommons.node.utils import (
     validate_hamiltonian,
     validate_shape,
@@ -46,7 +44,7 @@ from qctrlcommons.preconditions import (
 )
 
 
-class SparsePwcOperator(CoreNode):
+class SparsePwcOperator(Node):
     r"""
     Creates a sparse piecewise-constant operator (sparse-matrix-valued function of time).
 
@@ -68,12 +66,14 @@ class SparsePwcOperator(CoreNode):
 
     See Also
     --------
-    constant_sparse_pwc_operator : A function to create constant `SparsePwc`s.
-    sparse_pwc_sum : A function to calculate the sum of multiple `SparsePwc`s.
+    constant_sparse_pwc_operator : Create constant `SparsePwc`s.
+    density_matrix_evolution_pwc : Evolve a quantum state in an open system.
+    pwc_operator : Corresponding operation for `TensorPwc`s.
+    sparse_pwc_sum : Sum multiple `SparsePwc`s.
+    state_evolution_pwc : Evolve a quantum state.
     """
 
     name = "sparse_pwc_operator"
-    _module_attr = "create_sparse_pwc_operator"
     args = [
         forge.arg("signal", type=types.TensorPwc),
         forge.arg("operator", type=Union[np.ndarray, coo_matrix]),
@@ -109,7 +109,7 @@ class SparsePwcOperator(CoreNode):
         return SparsePwcNodeData(_operation, values_shape=operator_shape)
 
 
-class ConstantSparsePwcOperator(CoreNode):
+class ConstantSparsePwcOperator(Node):
     r"""
     Creates a constant sparse piecewise-constant operator over a specified duration.
 
@@ -128,12 +128,12 @@ class ConstantSparsePwcOperator(CoreNode):
 
     See Also
     --------
-    sparse_pwc_operator : A function to create time-dependent sparse piecewise-constant operators.
-    sparse_pwc_sum : A function to calculate the sum of multiple `SparsePwc`s.
+    constant_pwc_operator : Corresponding operation for `TensorPwc`s.
+    sparse_pwc_operator : Create `SparsePwc` operators.
+    sparse_pwc_sum : Sum multiple `SparsePwc`s.
     """
 
     name = "constant_sparse_pwc_operator"
-    _module_attr = "create_constant_sparse_pwc_operator"
     args = [
         forge.arg("duration", type=float),
         forge.arg("operator", type=Union[np.ndarray, coo_matrix]),
@@ -156,7 +156,7 @@ class ConstantSparsePwcOperator(CoreNode):
         return SparsePwcNodeData(_operation, values_shape=operator_shape)
 
 
-class SparsePwcSum(CoreNode):
+class SparsePwcSum(Node):
     r"""
     Creates the sum of multiple sparse-matrix-valued piecewise-constant functions.
 
@@ -173,10 +173,15 @@ class SparsePwcSum(CoreNode):
     SparsePwc
         The piecewise-constant function of time :math:`\sum_j v_j(t)`. It
         has the same shape as each of the `terms` that you provided.
+
+    See Also
+    --------
+    constant_sparse_pwc_operator : Create constant `SparsePwc`s.
+    pwc_sum : Corresponding operation for `TensorPwc`s.
+    sparse_pwc_operator : Create `SparsePwc` operators.
     """
 
     name = "sparse_pwc_sum"
-    _module_attr = "get_sparse_pwc_sum"
     args = [forge.arg("terms", type=List[types.SparsePwc])]
     kwargs = {}  # SparsePwcSum doesn't accept `name` as an argument.
     rtype = types.SparsePwc
@@ -199,7 +204,7 @@ class SparsePwcSum(CoreNode):
         return SparsePwcNodeData(_operation, values_shape=shape_0)
 
 
-class StateEvolutionPwc(CoreNode):
+class StateEvolutionPwc(Node):
     r"""
     Calculates the time evolution of a state generated by a piecewise-constant
     Hamiltonian by using the Lanczos method.
@@ -208,7 +213,7 @@ class StateEvolutionPwc(CoreNode):
     ----------
     initial_state : Union[Tensor, np.ndarray]
         The initial state as a Tensor or np.ndarray of shape ``[D]``.
-    hamiltonian : Union[PwcOperator, SparsePwc]
+    hamiltonian : Union[TensorPwc, SparsePwc]
         The control Hamiltonian. Uses sparse matrix multiplication if of type
         `SparsePwc`, which can be more efficient for large operators that are
         relatively sparse (contain mostly zeros).
@@ -230,6 +235,30 @@ class StateEvolutionPwc(CoreNode):
         dimension) represents the state at ``sample_times[n]`` evolved from
         the initial state.
 
+    Warnings
+    --------
+    This calculation can be relatively inefficient for small systems (very roughly
+    speaking, when the dimension of your Hilbert space is less than around 100; the
+    exact cutoff depends on the specifics of your problem though). You should generally
+    first try using `time_evolution_operators_pwc` to get the full time evolution
+    operators (and evolve your state using those), and only switch to this method
+    if that approach proves too slow or memory intensive. See the
+    `Simulate quantum dynamics user guide`_ for an example of calculating state
+    evolution with `time_evolution_operators_pwc`.
+
+    .. _Simulate quantum dynamics user guide:
+        https://docs.q-ctrl.com/boulder-opal/user-guides/simulate-quantum-dynamics#Example-Generating-entangled-states-in-atomic-chains
+
+    See Also
+    --------
+    density_matrix_evolution_pwc : Corresponding operation for open systems.
+    discretize_stf : Discretize an `Stf` into a `TensorPwc`.
+    estimated_krylov_subspace_dimension_lanczos : Obtain a Krylov subspace dimension to use
+        with this integrator.
+    sparse_pwc_operator : Create `SparsePwc` operators.
+    time_evolution_operators_pwc : Unitary time evolution operators for quantum systems with
+        `TensorPwc` Hamiltonians.
+
     Notes
     -----
     The Lanczos algorithm calculates the unitary evolution of a state in the Krylov
@@ -241,7 +270,6 @@ class StateEvolutionPwc(CoreNode):
     """
 
     name = "state_evolution_pwc"
-    _module_attr = "propagate_state_lanczos"
     args = [
         forge.arg("initial_state", type=Union[types.Tensor, np.ndarray]),
         forge.arg("hamiltonian", type=Union[types.TensorPwc, types.SparsePwc]),
@@ -282,15 +310,62 @@ class StateEvolutionPwc(CoreNode):
         return TensorNodeData(_operation, shape=shape)
 
 
-class EstimatedKrylovSubspaceDimension(CoreNode):
+class EstimatedKrylovSubspaceDimension(Node):
+    r"""
+    This node will be removed in the future.
+    Please use :py:func:`estimated_krylov_subspace_dimension_lanczos` instead.
     """
-    Estimates the dimension of the Krylov subspace to be used in the Lanczos integrator for
-    a given bound on the total error in the integration.
 
-    You can use this function to find a proper dimension of the Krylov subspace for state
-    propagation using the Lanczos algorithm. Note that you can provide your own estimation
-    of the Hamiltonian spectral range or use the `spectral_range` operation to
-    perform that calculation.
+    name = "estimated_krylov_subspace_dimension"
+    args = [
+        forge.arg("spectral_range", type=Union[types.Tensor, float]),
+        forge.arg("duration", type=float),
+        forge.arg("maximum_segment_duration", type=float),
+        forge.arg("error_tolerance", type=float, default=1e-6),
+    ]
+    rtype = types.Tensor
+
+    @classmethod
+    def create_node_data(cls, _operation, **kwargs):
+        warnings.warn(
+            "The estimated_krylov_subspace_dimension node will be removed in the future. "
+            "Please use estimated_krylov_subspace_dimension_lanczos instead."
+        )
+        duration = kwargs.get("duration")
+        maximum_segment_duration = kwargs.get("maximum_segment_duration")
+        error_tolerance = kwargs.get("error_tolerance")
+        check_argument(
+            maximum_segment_duration > 0,
+            "The maximum segment duration must be positive.",
+            {"maximum_segment_duration": maximum_segment_duration},
+        )
+        check_argument(
+            duration > 0, "The duration must be positive.", {"duration": duration}
+        )
+        check_argument(
+            maximum_segment_duration <= duration,
+            "The maximum segment duration must be less than or equal to duration.",
+            {
+                "maximum_segment_duration": maximum_segment_duration,
+                "duration": duration,
+            },
+        )
+        check_argument(
+            error_tolerance > 0,
+            "The error tolerance must be positive.",
+            {"error_tolerance": error_tolerance},
+        )
+
+        return TensorNodeData(_operation, shape=())
+
+
+class EstimatedKrylovSubspaceDimensionLanczos(Node):
+    r"""
+    Calculates an appropriate Krylov subspace dimension (:math:`k`) to use in the Lanczos
+    integrator while keeping the total error in the evolution below a given error tolerance.
+
+    Note that you can provide your own estimation of the Hamiltonian spectral range or use the
+    `spectral_range` operation to perform that calculation.
 
     Parameters
     ----------
@@ -311,24 +386,45 @@ class EstimatedKrylovSubspaceDimension(CoreNode):
     Returns
     -------
     Tensor
-        Estimated dimension of the Krylov subspace for integration.
+        Recommended value of :math:`k` to use in a Lanczos integration with a Hamiltonian with a
+        similar spectral range to the one passed.
 
     See Also
     --------
-    spectral_range, state_evolution_pwc
+    spectral_range : Range of the eigenvalues of a Hermitian operator.
+    state_evolution_pwc : Evolve a quantum state.
+
+    Notes
+    -----
+    To provide the recommended :math:`k` parameter, this function uses the bound in the error for
+    the Lanczos algorithm [1]_ [2]_ as an estimate for the error. For a single time step this gives
+
+    .. math::
+        \mathrm{error} \leq 12 \exp \left( - \frac{(w\tau)^2}{16 k} \right)
+        \left (\frac{e w \tau}{ 4 k}  \right )^{k}
+
+    where :math:`\tau` is the time step and :math:`w` is the spectral range of the Hamiltonian.
+
+    As this bound overestimates the error, the actual resulting errors with the recommended
+    parameter are expected to be (a few orders of magnitude) smaller than the requested tolerance.
+
+    References
+    ----------
+    .. [1] `N. Del Buono and L. Lopez,
+           Lecture Notes in Computer Science 2658, 111 (2003).
+           <https://doi.org/10.1007/3-540-44862-4_13>`_
+
+    .. [2] `M. Hochbruck and C. Lubich,
+           SIAM Journal on Numerical Analysis 34, 1911 (1997).
+           <https://doi.org/10.1137/S0036142995280572>`_
     """
 
-    name = "estimated_krylov_subspace_dimension"
-    _module_attr = "recommend_lanczos_k"
+    name = "estimated_krylov_subspace_dimension_lanczos"
     args = [
         forge.arg("spectral_range", type=Union[types.Tensor, float]),
         forge.arg("duration", type=float),
         forge.arg("maximum_segment_duration", type=float),
-        forge.arg(
-            "error_tolerance",
-            type=float,
-            default=1e-6,
-        ),
+        forge.arg("error_tolerance", type=float, default=1e-6),
     ]
     rtype = types.Tensor
 
@@ -343,9 +439,7 @@ class EstimatedKrylovSubspaceDimension(CoreNode):
             {"maximum_segment_duration": maximum_segment_duration},
         )
         check_argument(
-            duration > 0,
-            "The duration must be positive.",
-            {"duration": duration},
+            duration > 0, "The duration must be positive.", {"duration": duration}
         )
         check_argument(
             maximum_segment_duration <= duration,
@@ -358,15 +452,13 @@ class EstimatedKrylovSubspaceDimension(CoreNode):
         check_argument(
             error_tolerance > 0,
             "The error tolerance must be positive.",
-            {
-                "error_tolerance": error_tolerance,
-            },
+            {"error_tolerance": error_tolerance},
         )
 
         return TensorNodeData(_operation, shape=())
 
 
-class SpectralRange(CoreNode):
+class SpectralRange(Node):
     r"""
     Obtains the range of the eigenvalues of a Hermitian operator.
 
@@ -428,7 +520,6 @@ class SpectralRange(CoreNode):
     """
 
     name = "spectral_range"
-    _module_attr = "calculate_spectral_range"
     args = [
         forge.arg("operator", type=Union[types.Tensor, np.ndarray, coo_matrix]),
         forge.arg("iteration_count", type=int, default=3000),
@@ -455,8 +546,5 @@ class SpectralRange(CoreNode):
             {"iteration_count": iteration_count},
         )
         if not isinstance(operator, TensorNodeData):
-            if issparse(operator):
-                check_argument_hermitian(operator.toarray(), "operator")
-            else:
-                check_argument_hermitian(operator, "operator")
+            check_argument_hermitian(operator, "operator")
         return TensorNodeData(_operation, shape=())
