@@ -22,25 +22,193 @@ import forge
 import numpy as np
 
 from qctrlcommons.node import types
-from qctrlcommons.node.module import CoreNode
+from qctrlcommons.node.base import Node
 from qctrlcommons.node.node_data import (
     StfNodeData,
+    TensorNodeData,
     TensorPwcNodeData,
 )
-from qctrlcommons.node.tensorflow import TensorNodeData
 from qctrlcommons.node.utils import (
+    validate_batch_and_values_shapes,
     validate_hamiltonian,
-    validate_values_and_batch_shape,
+    validate_shape,
 )
 from qctrlcommons.preconditions import (
     check_argument,
     check_argument_integer,
     check_argument_iteratable,
+    check_argument_numeric,
+    check_operator,
     check_sample_times,
 )
 
 
-class StfSum(CoreNode):
+class StfOperator(Node):
+    """
+    Creates a constant operator multiplied by a sampleable signal.
+
+    Parameters
+    ----------
+    signal : Stf
+        A sampleable function representing the signal :math:`a(t)`
+        or a batch of sampleable functions.
+    operator : np.ndarray or Tensor
+        The operator :math:`A`. It must have two equal dimensions.
+
+    Returns
+    -------
+    Stf
+        The sampleable operator :math:`a(t)A` (or batch of sampleable operators, if
+        you provide a batch of signals).
+
+    See Also
+    --------
+    constant_stf_operator : Create a constant `Stf` operator.
+    pwc_operator : Corresponding operation for `TensorPwc`s.
+    stf_sum : Sum multiple `Stf`s.
+    """
+
+    name = "stf_operator"
+    args = [
+        forge.arg("signal", type=types.Stf),
+        forge.arg("operator", type=Union[np.ndarray, types.Tensor]),
+    ]
+    kwargs = {}  # Stfs don't accept name as an argument.
+    rtype = types.Stf
+
+    @classmethod
+    def create_node_data(cls, _operation, **kwargs):
+        signal = kwargs.get("signal")
+        operator = kwargs.get("operator")
+        check_argument(
+            isinstance(signal, StfNodeData),
+            "The signal must be an Stf.",
+            {"signal": signal},
+        )
+        batch_shape, _ = validate_batch_and_values_shapes(signal, "signal")
+        values_shape = validate_shape(operator, "operator")
+        check_operator(operator, "operator")
+        check_argument(
+            len(values_shape) == 2,
+            "The operator must be a matrix, not a batch.",
+            {"operator": operator},
+            extras={"operator.shape": values_shape},
+        )
+        return StfNodeData(
+            _operation,
+            values_shape=values_shape,
+            batch_shape=batch_shape,
+        )
+
+
+class ConstantStfOperator(Node):
+    r"""
+    Creates a constant operator.
+
+    Parameters
+    ----------
+    operator : np.ndarray or Tensor
+        The operator :math:`A`, or a batch of operators. It must have at
+        least two dimensions, and its last two dimensions must be equal.
+
+    Returns
+    -------
+    Stf(3D)
+        The operator :math:`t\mapsto A` (or batch of
+        operators, if you provide a batch of operators).
+
+    See Also
+    --------
+    constant_pwc_operator : Corresponding operation for `TensorPwc`s.
+    constant_stf: Create a batch of constant `Stf`s.
+    stf_operator : Create an `Stf` operator.
+    """
+
+    name = "constant_stf_operator"
+    args = [forge.arg("operator", type=Union[np.ndarray, types.Tensor])]
+    kwargs = {}  # Stfs don't accept name as an argument.
+    rtype = types.Stf
+
+    @classmethod
+    def create_node_data(cls, _operation, **kwargs):
+        operator = kwargs.get("operator")
+        shape = validate_shape(operator, "operator")
+        check_operator(operator, "operator")
+        return StfNodeData(
+            _operation,
+            values_shape=shape[-2:],
+            batch_shape=shape[:-2],
+        )
+
+
+class ConstantStf(Node):
+    r"""
+    Create a constant sampleable tensor-valued function of time.
+
+    Parameters
+    ----------
+    constant : np.ndarray or Tensor
+        The constant value :math:`c` of the function.
+        To create a batch of :math:`B_1 \times \ldots \times B_n` constant
+        functions of shape :math:`D_1 \times \ldots \times D_m`, provide this `constant`
+        parameter as an object of shape
+        :math:`B_1\times\ldots\times B_n\times D_1\times\ldots\times D_m`.
+    batch_dimension_count : int, optional
+        The number of batch dimensions, :math:`n`, in `constant`.
+        If provided, the first :math:`n` dimensions of `constant` are considered batch dimensions.
+        Defaults to 0, which corresponds to no batch.
+
+    Returns
+    -------
+    Stf
+       An Stf representing the constant function :math:`f(t) = c` for all time
+       :math:`t` (or a batch of functions, if you provide `batch_dimension_count`).
+
+    See Also
+    --------
+    constant_stf_operator : Create a constant sampleable function from operators.
+    identity_stf : Create an `Stf` representing the identity function.
+    constant_pwc : Corresponding operation for `TensorPwc`s.
+    """
+
+    name = "constant_stf"
+    args = [
+        forge.arg("constant", type=Union[np.ndarray, types.Tensor]),
+        forge.arg("batch_dimension_count", type=int, default=0),
+    ]
+    kwargs = {}  # Stfs don't accept name as an argument.
+    rtype = types.Stf
+
+    @classmethod
+    def create_node_data(cls, _operation, **kwargs):
+        constant = kwargs.get("constant")
+        batch_dimension_count = kwargs.get("batch_dimension_count")
+
+        check_argument_numeric(constant, "constant")
+        check_argument_integer(batch_dimension_count, "batch_dimension_count")
+
+        shape = validate_shape(constant, "constant")
+        check_argument(
+            len(shape) >= batch_dimension_count,
+            "The number of batch dimensions must not be larger than the number of "
+            "dimensions of the input constant.",
+            {"constant": constant, "batch_dimension_count": batch_dimension_count},
+            {"Number of value dimensions": len(constant.shape)},
+        )
+        check_argument(
+            batch_dimension_count >= 0,
+            "The number of batch dimensions must not be negative.",
+            {"batch_dimension_count": batch_dimension_count},
+        )
+
+        return StfNodeData(
+            operation=_operation,
+            values_shape=shape[batch_dimension_count:],
+            batch_shape=shape[:batch_dimension_count],
+        )
+
+
+class StfSum(Node):
     r"""
     Creates the sum of multiple sampleable functions.
 
@@ -54,10 +222,15 @@ class StfSum(CoreNode):
     Stf
         The sampleable function of time :math:`\sum_j v_j(t)`. It has the same
         shape as each of the `terms` that you provide.
+
+    See Also
+    --------
+    pwc_sum : Corresponding operation for `TensorPwc`s.
+    stf_operator : Create an `Stf` operator.
+    stf_sum : Sum multiple `Stf`s.
     """
 
     name = "stf_sum"
-    _module_attr = "get_stf_sum"
     args = [forge.arg("terms", type=List[types.Stf])]
     kwargs = {}  # Stfs don't accept name as an argument.
     rtype = types.Stf
@@ -67,23 +240,21 @@ class StfSum(CoreNode):
         terms = kwargs.get("terms")
         check_argument_iteratable(terms, "terms")
         check_argument(
-            all([isinstance(term, StfNodeData) for term in terms]),
+            all(isinstance(term, StfNodeData) for term in terms),
             "Each of the terms must be an Stf.",
             {"terms": terms},
         )
-        values_shape, batch_shape = validate_values_and_batch_shape(
+        batch_shape, values_shape = validate_batch_and_values_shapes(
             terms[0],
             "terms[0]",
         )
         check_argument(
             all(
-                [
-                    (
-                        (values_shape == term.values_shape)
-                        and (batch_shape == term.batch_shape)
-                    )
-                    for term in terms[1:]
-                ]
+                (
+                    (values_shape == term.values_shape)
+                    and (batch_shape == term.batch_shape)
+                )
+                for term in terms[1:]
             ),
             "All the terms must have the same shape.",
             {"terms": terms},
@@ -95,7 +266,7 @@ class StfSum(CoreNode):
         )
 
 
-class DiscretizeStf(CoreNode):
+class DiscretizeStf(Node):
     r"""
     Creates a piecewise-constant function by discretizing a sampleable function.
 
@@ -144,34 +315,46 @@ class DiscretizeStf(CoreNode):
     .. math::
         w_n = \frac{1}{M}
         \sum_{m=1}^M v\left(t_{n-1} + \left(m-\tfrac{1}{2}\right) \frac{\tau}{MN} \right).
+
+    See Also
+    --------
+    convolve_pwc : Create an `Stf` by convolving a `TensorPwc` with a kernel.
+    identity_stf : Create an `Stf` representing the identity function.
+    sample_stf : Sample an `Stf` at given times.
     """
 
     name = "discretize_stf"
-    _module_attr = "discretize_stf"
     args = [
         forge.arg("stf", type=types.Stf),
         forge.arg("duration", type=float),
         forge.arg("segments_count", type=int),
         forge.arg("samples_per_segment", type=int, default=1),
     ]
-    rtype = types.Tensor
+    rtype = types.TensorPwc
 
     @classmethod
     def create_node_data(cls, _operation, **kwargs):
         stf = kwargs.get("stf")
         duration = kwargs.get("duration")
         segments_count = kwargs.get("segments_count")
-        values_shape, batch_shape = validate_values_and_batch_shape(
+        batch_shape, values_shape = validate_batch_and_values_shapes(
             stf,
             "stf",
         )
         check_argument_integer(segments_count, "segments_count")
         check_argument(
             segments_count > 0,
-            "The segments count must be greater than zero.",
+            "The number of segments must be greater than zero.",
             {"segments_count": segments_count},
         )
         durations = duration / segments_count * np.ones(segments_count)
+        samples_per_segment = kwargs.get("samples_per_segment")
+        check_argument_integer(samples_per_segment, "samples_per_segment")
+        check_argument(
+            samples_per_segment > 0,
+            "The number of samples per segment to take must be greater than zero.",
+            {"samples_per_segment": samples_per_segment},
+        )
         return TensorPwcNodeData(
             _operation,
             values_shape=values_shape,
@@ -180,7 +363,7 @@ class DiscretizeStf(CoreNode):
         )
 
 
-class TimeEvolutionOperatorsStf(CoreNode):
+class TimeEvolutionOperatorsStf(Node):
     """
     Calculates the time-evolution operators for a system defined by an STF Hamiltonian by using a
     4th order Runge–Kutta method.
@@ -206,10 +389,13 @@ class TimeEvolutionOperatorsStf(CoreNode):
         Tensor of shape [..., N, D, D], representing the unitary time evolution.
         The n-th element (along the -3 dimension) represents the unitary (or batch of unitaries)
         from t = 0 to ``sample_times[n]``.
+
+    See Also
+    --------
+    time_evolution_operators_pwc : Corresponding operation for `TensorPwc` Hamiltonians.
     """
 
     name = "time_evolution_operators_stf"
-    _module_attr = "create_unitaries_from_stf_hamiltonian"
     args = [
         forge.arg("hamiltonian", type=types.Stf),
         forge.arg("sample_times", type=np.ndarray),
@@ -227,7 +413,7 @@ class TimeEvolutionOperatorsStf(CoreNode):
             "The Hamiltonian must be an Stf.",
             {"hamiltonian": hamiltonian},
         )
-        values_shape, batch_shape = validate_values_and_batch_shape(
+        batch_shape, values_shape = validate_batch_and_values_shapes(
             hamiltonian,
             "hamiltonian",
         )
@@ -252,7 +438,7 @@ class TimeEvolutionOperatorsStf(CoreNode):
         return TensorNodeData(_operation, shape=shape)
 
 
-class ConvolvePwc(CoreNode):
+class ConvolvePwc(Node):
     r"""
     Creates the convolution of a piecewise-constant function with a kernel.
 
@@ -284,10 +470,18 @@ class ConvolvePwc(CoreNode):
     frequency domain, so this function can be viewed as applying a linear
     time-invariant filter (specified via its time domain kernel :math:`K(t)`)
     to :math:`\alpha(t)`.
+
+    See Also
+    --------
+    discretize_stf : Discretize an `Stf` into a `TensorPwc`.
+    gaussian_integral_function : Create a function that computes the integral of a normalized
+        Gaussian.
+    sample_stf : Sample an `Stf` at given times.
+    sinc_integral_function : Create a function that computes the integral of the sinc function.
+    tensor_pwc : Create piecewise-constant functions.
     """
 
     name = "convolve_pwc"
-    _module_attr = "convolve_pwc"
     args = [
         forge.arg("pwc", type=types.TensorPwc),
         forge.arg("kernel_integral", type=Callable),
@@ -298,7 +492,7 @@ class ConvolvePwc(CoreNode):
     @classmethod
     def create_node_data(cls, _operation, **kwargs):
         pwc = kwargs.get("pwc")
-        values_shape, batch_shape = validate_values_and_batch_shape(pwc, "pwc")
+        batch_shape, values_shape = validate_batch_and_values_shapes(pwc, "pwc")
         return StfNodeData(
             _operation,
             values_shape=values_shape,
@@ -306,7 +500,7 @@ class ConvolvePwc(CoreNode):
         )
 
 
-class SincIntegralFunction(CoreNode):
+class SincIntegralFunction(Node):
     r"""
     Creates a function that computes the integral of the sinc function.
 
@@ -341,7 +535,6 @@ class SincIntegralFunction(CoreNode):
     """
 
     name = "sinc_integral_function"
-    _module_attr = "create_sinc_integral_function"
     args = [
         forge.arg("cut_off_frequency", type=Union[float, types.Tensor]),
     ]
@@ -349,7 +542,7 @@ class SincIntegralFunction(CoreNode):
     rtype = types.Function
 
 
-class GaussianIntegralFunction(CoreNode):
+class GaussianIntegralFunction(Node):
     r"""
     Creates a function that computes the integral of a normalized Gaussian.
 
@@ -388,10 +581,86 @@ class GaussianIntegralFunction(CoreNode):
     """
 
     name = "gaussian_integral_function"
-    _module_attr = "create_gaussian_integral_function"
     args = [
         forge.arg("std", type=Union[float, types.Tensor]),
         forge.arg("offset", type=Optional[Union[float, types.Tensor]], default=0),
     ]
     kwargs = {}  # Functions don't accept name as an argument.
     rtype = types.Function
+
+
+class SampleStf(Node):
+    """
+    Samples an Stf at the given times.
+
+    Parameters
+    ----------
+    stf : Stf
+        The Stf to sample.
+    sample_times : np.ndarray(1D, real)
+        The times at which you want to sample the Stf. Must be ordered and contain
+        at least one element.
+    name : str, optional
+        The name of the node.
+
+    Returns
+    -------
+    Tensor
+        The values of the Stf at the given times.
+
+    See Also
+    --------
+    constant_stf_operator : Create a constant `Stf` operator.
+    discretize_stf : Discretize an `Stf` into a `TensorPwc`.
+    identity_stf : Create an `Stf` representing the identity function.
+    """
+
+    name = "sample_stf"
+    args = [
+        forge.arg("stf", type=types.Stf),
+        forge.arg("sample_times", type=np.ndarray),
+    ]
+    rtype = types.Tensor
+
+    @classmethod
+    def create_node_data(cls, _operation, **kwargs):
+        stf = kwargs.get("stf")
+        check_argument(
+            isinstance(stf, StfNodeData), "The stf must be an Stf.", {"stf": stf}
+        )
+        batch_shape, values_shape = validate_batch_and_values_shapes(stf, "stf")
+        sample_times = kwargs.get("sample_times")
+        check_sample_times(sample_times, "sample_times")
+        time_count = len(sample_times)
+        shape = batch_shape + (time_count,) + values_shape
+        return TensorNodeData(_operation, shape=shape)
+
+
+class IdentityStf(Node):
+    """
+    Returns an Stf representing the identity function, f(t) = t.
+
+    Returns
+    -------
+    Stf
+        An Stf representing the identity function.
+
+    See Also
+    --------
+    constant_stf: Create a batch of constant `Stf`s.
+    discretize_stf : Discretize an `Stf` into a `TensorPwc`.
+    sample_stf : Sample an `Stf` at given times.
+    """
+
+    name = "identity_stf"
+    args = []
+    kwargs = {}  # Stfs don't accept name as an argument.
+    rtype = types.Stf
+
+    @classmethod
+    def create_node_data(cls, _operation, **kwargs):
+        return StfNodeData(
+            _operation,
+            values_shape=(),
+            batch_shape=(),
+        )
